###############################################################################
#                                                                             #
# Peekaboo Extended Email Attachment Behavior Observation Owl                 #
#                                                                             #
# daemon.py                                                                   #
###############################################################################
#                                                                             #
# Copyright (C) 2016-2020  science + computing ag                             #
#                                                                             #
# This program is free software: you can redistribute it and/or modify        #
# it under the terms of the GNU General Public License as published by        #
# the Free Software Foundation, either version 3 of the License, or (at       #
# your option) any later version.                                             #
#                                                                             #
# This program is distributed in the hope that it will be useful, but         #
# WITHOUT ANY WARRANTY; without even the implied warranty of                  #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU           #
# General Public License for more details.                                    #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#                                                                             #
###############################################################################

""" The main peekaboo module, starting up and managing all the various
components. """

import errno
import gettext
import os
import sys
import grp
import pwd
import logging
import signal
import socket
from argparse import ArgumentParser
from sdnotify import SystemdNotifier
from sqlalchemy.exc import SQLAlchemyError
from peekaboo import PEEKABOO_OWL, __version__
from peekaboo.config import PeekabooConfig, PeekabooConfigParser
from peekaboo.db import PeekabooDatabase
from peekaboo.queuing import JobQueue
from peekaboo.ruleset.engine import RulesetEngine
from peekaboo.sample import SampleFactory
from peekaboo.server import PeekabooServer
from peekaboo.exceptions import PeekabooDatabaseError, \
        PeekabooConfigException, PeekabooRulesetConfigError
from peekaboo.toolbox.cuckoo import CuckooEmbed, CuckooApi


logger = logging.getLogger(__name__)


class SignalHandler():
    """ Signal handler. """
    def __init__(self):
        """ register custom signal handler """
        self.listeners = []

        signal.signal(signal.SIGINT, self.signal_handler)
        signal.signal(signal.SIGTERM, self.signal_handler)
        signal.signal(signal.SIGCHLD, self.signal_handler)

    def register_listener(self, listener):
        """ Register a listener object which is expected to implement a very
        simple interface: Method shut_down() is called if SIGINT or SIGTERM
        arrive, reap_children() is called if SIGCHLD arrives. Both are expected
        to defer actual handling of the condition. """
        self.listeners.append(listener)

    def signal_handler(self, sig, frame):
        """ catch signal and call appropriate methods in registered listener
        classes """
        if sig == signal.SIGINT or sig == signal.SIGTERM:
            logger.debug("SIGINT/TERM")

            # these should take serious care about being called across threads
            for listener in self.listeners:
                listener.shut_down()

        if sig == signal.SIGCHLD:
            logger.debug("SIGCHLD")
            for listener in self.listeners:
                listener.reap_children()


class PeekabooDaemonInfrastructure(object):
    """ A class that manages typical daemon infrastructure such as PID file and
    privileges. """
    def __init__(self, pid_file, sock_file, user, group):
        self.pid_file = pid_file
        self.sock_file = sock_file
        self.user = user
        self.group = group

        self.pid_file_created = False

    def init(self):
        """ Initialize daemon infrastructure. """
        self.drop_privileges()
        self.create_pid_file()
        self.check_stale_socket()

    def drop_privileges(self):
        """ Check and potentially drop privileges. """
        if os.getuid() != 0:
            return

        if not self.user:
            logger.warning('Peekaboo should not run as root. Please '
                           'configure a user to run as.')
            sys.exit(1)

        # drop privileges to user
        try:
            userdata = pwd.getpwnam(self.user)
        except KeyError as notfound:
            logger.critical('Error looking up daemon user: %s', notfound)
            sys.exit(1)

        uid = userdata[2]
        gid = userdata[3]
        userhome = userdata[5]

        gid_log = ''
        if self.group:
            try:
                gid = grp.getgrnam(self.group)[2]
            except KeyError as notfound:
                logger.critical('Error looking up daemon group: %s', notfound)
                sys.exit(1)
            gid_log = ' and group %s' % self.group

        os.initgroups(self.user, gid)
        os.setgid(gid)
        os.setuid(uid)

        grouplist = []
        for gid in os.getgroups():
            groupdata = grp.getgrgid(gid)
            grouplist.append('%s(%d)' % (groupdata[0], groupdata[2]))

        logger.info('After dropping privileges to user %s%s now running as '
                    'user %s(%d) with primary group %s(%d) and supplementary '
                    'groups %s', self.user, gid_log,
                    pwd.getpwuid(os.getuid())[0], os.getuid(),
                    grp.getgrgid(os.getgid())[0], os.getgid(),
                    ', '.join(grouplist))

        # set $HOME to the users home directory
        # (VirtualBox must access the configs)
        os.environ['HOME'] = userhome
        logger.debug('$HOME is %s', os.environ['HOME'])

    def create_pid_file(self):
        """ Check for stale old and create a new PID file. Look at the socket
        as well. """
        pid = None
        if os.path.exists(self.pid_file):
            stale = False
            try:
                with open(self.pid_file, 'r') as pidfile:
                    pid = int(pidfile.read())
            except (OSError, IOError, ValueError) as error:
                stale = True
                logger.warning('PID file exists but cannot be read, '
                               'assuming it to be stale')

            if pid is not None:
                try:
                    # ping the process to see if it exists, sends no signal
                    os.kill(pid, 0)
                except OSError as oserror:
                    # ESRCH == no such process
                    if oserror.errno == errno.ESRCH:
                        stale = True

            if not stale:
                logger.critical('Another instance of Peekaboo seems to be '
                                'running as process %d. Please check PID '
                                'file %s.', pid, self.pid_file)
                sys.exit(1)

            logger.warning('Removing stale PID file of process %d', pid)
            try:
                os.remove(self.pid_file)
            except OSError as error:
                logger.critical('Error deleting stale PID file %s: %s',
                                self.pid_file, error)
                sys.exit(1)

        # write PID file
        pid = os.getpid()
        with open(self.pid_file, "w") as pidfile:
            pidfile.write("%d\n" % pid)

        # remember that the PID file is ours - important on shutdown
        self.pid_file_created = True
        logger.debug('PID %d written to %s', pid, self.pid_file)

    def check_stale_socket(self):
        """ Check if the socket file exists already/still and if it is stale or
        actively serviced. Remove it if stale. """
        # is the socket also stale?
        if not os.path.exists(self.sock_file):
            return

        stale = False
        try:
            sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
            sock.connect(self.sock_file)
            logger.debug('Someone answered on existing socket')
        except socket.error as sockerr:
            logger.debug('Existing socket connection attempt failed: %s',
                         sockerr)
            if sockerr.errno == errno.ECONNREFUSED:
                stale = True

        if not stale:
            logger.critical('Socket %s exists and seems to be serviced. '
                            'Please check for another instance running.',
                            self.sock_file)
            sys.exit(1)

        logger.warning('Removing stale socket %s', self.sock_file)
        try:
            os.remove(self.sock_file)
        except OSError as oserror:
            logger.critical('Error removing stale socket %s: %s',
                            self.sock_file, oserror)
            sys.exit(1)

    def __del__(self):
        """ Clean up on shutdown, such as removing the PID file. """
        # only remove stuff if we created it. Otherwise we're bailing (but
        # still getting called) after realising that another instance is
        # running.
        if not self.pid_file_created:
            return

        logger.debug('Removing PID file %s', self.pid_file)
        try:
            os.remove(self.pid_file)
        except OSError as oserror:
            logger.warning('Removal of PID file %s failed: %s',
                           self.pid_file, oserror)


def run():
    """ Runs the Peekaboo daemon. """
    arg_parser = ArgumentParser(
        description='Peekaboo Extended Email Attachment Behavior Observation Owl'
    )
    arg_parser.add_argument(
        '-c', '--config',
        action='store',
        help='The configuration file for Peekaboo.'
    )
    arg_parser.add_argument(
        '-d', '--debug',
        action='store_true',
        help="Run Peekaboo in debug mode regardless of what's specified in the configuration."
    )
    arg_parser.add_argument(
        '-D', '--daemon',
        action='store_true',
        help='Run Peekaboo in daemon mode (suppresses the logo to be written to STDOUT).'
    )
    args = arg_parser.parse_args()

    print('Starting Peekaboo %s.' % __version__)
    if not args.daemon:
        print(PEEKABOO_OWL)

    # Check if CLI arguments override the configuration
    log_level = None
    if args.debug:
        log_level = logging.DEBUG

    try:
        config = PeekabooConfig(config_file=args.config, log_level=log_level)
        logger.debug(config)
    except PeekabooConfigException as error:
        logging.critical(error)
        sys.exit(1)

    if sys.version_info[0] < 3:
        logger.warning(
            "Python 2 support is deprecated and will be removed in a future "
            "release. Please switch to python 3.")

    # find localisation in our package directory
    locale_domain = 'peekaboo'
    locale_dir = os.path.join(os.path.dirname(__file__), 'locale')
    languages = None
    if config.report_locale:
        logger.debug('Looking for translations for preconfigured locale "%s"',
                     config.report_locale)
        languages = [config.report_locale]
        if not gettext.find(locale_domain, locale_dir, languages):
            logger.warning('Translation file not found - falling back to '
                           'system configuration.')
            languages = None

    logger.debug('Installing report message translations')
    translation = gettext.translation(locale_domain, locale_dir, languages,
                                      fallback=True)
    # python2's gettext needs to be told explicitly to return unicode strings
    loc_kwargs = {}
    if sys.version_info[0] < 3:
        loc_kwargs = {'unicode': True}
    translation.install(loc_kwargs)

    # establish a connection to the database
    try:
        db_con = PeekabooDatabase(
            db_url=config.db_url, instance_id=config.cluster_instance_id,
            stale_in_flight_threshold=config.cluster_stale_in_flight_threshold,
            log_level=config.db_log_level)
    except PeekabooDatabaseError as error:
        logging.critical(error)
        sys.exit(1)
    except SQLAlchemyError as dberr:
        logger.critical('Failed to establish a connection to the database '
                        'at %s: %s', config.db_url, dberr)
        sys.exit(1)

    # initialize the daemon infrastructure such as PID file and dropping
    # privileges, automatically cleans up after itself when going out of scope
    daemon_infrastructure = PeekabooDaemonInfrastructure(
        config.pid_file, config.sock_file, config.user, config.group)
    daemon_infrastructure.init()

    systemd = SystemdNotifier()

    # clear all our in flight samples and all instances' stale in flight
    # samples
    db_con.clear_in_flight_samples()
    db_con.clear_stale_in_flight_samples()

    # a cluster duplicate interval of 0 disables the handler thread which is
    # what we want if we don't have an instance_id and therefore are alone
    cldup_check_interval = 0
    if config.cluster_instance_id > 0:
        cldup_check_interval = config.cluster_duplicate_check_interval
        if cldup_check_interval < 5:
            cldup_check_interval = 5
            logger.warning("Raising excessively low cluster duplicate check "
                           "interval to %d seconds.",
                           cldup_check_interval)

    # read in the ruleset configuration
    try:
        ruleset_config = PeekabooConfigParser(config.ruleset_config)
    except PeekabooConfigException as error:
        logging.critical(error)
        sys.exit(1)

    # create a single ruleset engine for all workers, instantiates all the
    # rules based on the ruleset configuration, is otherwise stateless to avoid
    # concurrent use by multiple worker threads
    try:
        engine = RulesetEngine(ruleset_config, db_con)
    except (KeyError, ValueError, PeekabooConfigException) as error:
        logging.critical('Ruleset configuration error: %s', error)
        sys.exit(1)
    except PeekabooRulesetConfigError as error:
        logging.critical(error)
        sys.exit(1)

    job_queue = JobQueue(
        worker_count=config.worker_count, ruleset_engine=engine,
        db_con=db_con,
        cluster_duplicate_check_interval=cldup_check_interval)

    if config.cuckoo_mode == "embed":
        logger.warning(
            "Embedded mode for Cuckoo is deprecated and will be removed in "
            "a future release. Please switch to REST API mode.")

        cuckoo = CuckooEmbed(job_queue, config.cuckoo_exec,
                             config.cuckoo_submit, config.cuckoo_storage,
                             config.interpreter)
    # otherwise it's the new API method and default
    else:
        cuckoo = CuckooApi(job_queue, config.cuckoo_url,
                           config.cuckoo_api_token,
                           config.cuckoo_poll_interval,
                           config.cuckoo_submit_original_filename,
                           config.cuckoo_maximum_job_age)

    sig_handler = SignalHandler()
    sig_handler.register_listener(cuckoo)

    # Factory producing almost identical samples providing them with global
    # config values and references to other objects they need, such as cuckoo,
    # database connection and connection map.
    sample_factory = SampleFactory(
        cuckoo, config.sample_base_dir, config.job_hash_regex,
        config.keep_mail_data, config.processing_info_dir)

    # We only want to accept 2 * worker_count connections.
    try:
        server = PeekabooServer(
            sock_file=config.sock_file, job_queue=job_queue,
            sample_factory=sample_factory,
            request_queue_size=config.worker_count * 2,
            sock_group=config.sock_group,
            sock_mode=config.sock_mode)
    except Exception as error:
        logger.critical('Failed to start Peekaboo Server: %s', error)
        job_queue.shut_down()
        sys.exit(1)

    exit_code = 1
    try:
        systemd.notify("READY=1")
        # If this dies Peekaboo dies, since this is the main thread. (legacy)
        exit_code = cuckoo.do()
    except Exception as error:
        logger.critical('Main thread aborted: %s', error)
    finally:
        server.shutdown()
        job_queue.shut_down()
        try:
            db_con.clear_in_flight_samples()
            db_con.clear_stale_in_flight_samples()
        except PeekabooDatabaseError as dberr:
            logger.error(dberr)

    sys.exit(exit_code)

if __name__ == '__main__':
    run()



###############################################################################
#                                                                             #
# Peekaboo Extended Email Attachment Behavior Observation Owl                 #
#                                                                             #
# toolbox/                                                                    #
#         ole.py                                                           #
###############################################################################
#                                                                             #
# Copyright (C) 2016-2020  science + computing ag                             #
#                                                                             #
# This program is free software: you can redistribute it and/or modify        #
# it under the terms of the GNU General Public License as published by        #
# the Free Software Foundation, either version 3 of the License, or (at       #
# your option) any later version.                                             #
#                                                                             #
# This program is distributed in the hope that it will be useful, but         #
# WITHOUT ANY WARRANTY; without even the implied warranty of                  #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU           #
# General Public License for more details.                                    #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#                                                                             #
###############################################################################


import logging
import re
from oletools.olevba import VBA_Parser, FileOpenError
from oletools.olevba import detect_autoexec, detect_suspicious

logger = logging.getLogger(__name__)


class Oletools(object):
    """ Parent class, defines interface to Oletools. """
    def __init__(self, sample):
        self.sample = sample

    def get_report(self):
        """ Return oletools report or create if not already cached. """
        if self.sample.oletools_report is not None:
            return self.sample.oletools_report

        report = {
            'autoexec': [],
            'suspicious' : [],
        }

        file_path = self.sample.file_path
        try:
            vbaparser = VBA_Parser(file_path)

            # VBA_Parser reports macros for office documents
            report['has_macros'] = vbaparser.detect_vba_macros() or vbaparser.detect_xlm_macros()
            try:
                report['vba'] = vbaparser.reveal()
            except TypeError:
                # office document with no macros
                pass

            all_macros = vbaparser.extract_all_macros()
            if (report['has_macros'] and len(all_macros) == 1
                    and isinstance(all_macros[0], tuple)
                    and len(all_macros[0]) >= 3
                    and all_macros[0][2] == file_path):
                logger.warning(
                    "Buggy oletools version detected, result overridden. May "
                    "lead to false negatives, please update to fixed version")
                report['has_macros'] = False

            if vbaparser.detect_vba_macros():
                vb_code = vbaparser.extract_all_macros()
                for (_, _, _, c) in vb_code:
                    autoexec = detect_autoexec(c)
                    if len(autoexec) >= 1:
                        report['autoexec'].extend(autoexec)

                    suspicious = detect_suspicious(c)
                    if len(suspicious) >= 1:
                        report['suspicious'].extend(suspicious)

            vbaparser.close()
        except IOError:
            raise
        except (TypeError, FileOpenError):
            # The given file is not an office document.
            pass
        except Exception as error:
            logger.exception(error)

        report = OletoolsReport(report)
        self.sample.register_oletools_report(report)
        return report


class OletoolsReport(object):
    """ Represents a custom Oletools report. """
    def __init__(self, report=None):
        if report is None:
            report = {}
        self.report = report

    def __str__(self):
        return "<OletoolsReport('%s'>" % self.report

    @property
    def has_office_macros(self):
        """
        Detects macros in Microsoft Office documents.

        @return: True if macros where found, otherwise False.
                If VBA_Parser crashes it returns False too.
        """
        return self.report.get('has_macros', False)

    @property
    def vba_code(self):
        """
        Extracts vba code from Microsoft Office documents.
        @return: vba code if found, otherwise empty string.
        """
        return self.report.get('vba', '')

    @property
    def has_autoexec(self):
        """
        Uses olevba detect_autoexec and reports if something was found.
        @return: True or False
        """
        if self.report.get('autoexec', []):
            return True
        return False

    @property
    def detected_autoexec(self):
        """
        Method to access olevba detect_autoexec report.
        @return: String from List of Tuple(marker, explanation)
        """
        return "%s" % self.report.get('autoexec', [])

    @property
    def is_suspicious(self):
        """
        Uses olevba detect_suspicious and reports if something was found.
        @return: True or False
        """
        if self.report.get('suspicious', []):
            return True
        return False

    @property
    def detected_suspicious(self):
        """
        Method to access olevba detect_suspicious report.
        @return: String from List of Tuple(marker, explanation)
        """
        return "%s" % self.report.get('suspicious', [])

    def has_office_macros_with_suspicious_keyword(self, suspicious_keywords):
        """
        Detects macros with supplied suspicious keywords in Microsoft Office documents.

        @param suspicious_keywords: List of suspicious keyword regexes.
        @return: True if macros with keywords where found, otherwise False.
                If VBA_Parser crashes it returns False too.
        """
        vba = self.report.get('vba')
        if vba is None:
            return False

        suspicious = False
        for word in suspicious_keywords:
            if re.search(word, vba):
                suspicious = True
                break

        return suspicious



###############################################################################
#                                                                             #
# Peekaboo Extended Email Attachment Behavior Observation Owl                 #
#                                                                             #
# ruleset/                                                                    #
#         expressions.py                                                      #
###############################################################################
#                                                                             #
# Copyright (C) 2016-2020  science + computing ag                             #
# Based on pyparsing's eval_arith.py.
# Copyright 2009, 2011 Paul McGuire
#                                                                             #
# This program is free software: you can redistribute it and/or modify        #
# it under the terms of the GNU General Public License as published by        #
# the Free Software Foundation, either version 3 of the License, or (at       #
# your option) any later version.                                             #
#                                                                             #
# This program is distributed in the hope that it will be useful, but         #
# WITHOUT ANY WARRANTY; without even the implied warranty of                  #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU           #
# General Public License for more details.                                    #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#                                                                             #
###############################################################################

""" A simple expression grammar used for writing generic rules. """

from future.builtins import super

import logging
import operator
import re
from pyparsing import nums, alphas, alphanums, Word, Combine, Suppress, \
    oneOf, opAssoc, infixNotation, Literal, Keyword, Group, \
    delimitedList, QuotedString, ParserElement, ParseException
from peekaboo.ruleset import Result


logger = logging.getLogger(__name__)


class EvalBase(object):
    """ Base class of evaluatable objects providing common infrastructure. """
    def __init__(self, tokens):
        """ Just store the tokens for later evaluation. Expects all relevant
        tokens to be grouped together in the first element of the token list
        passed. This is the default for operand+operator+operand+... constructs
        with infixNotation and can be forced for others using Group():

        rvallist = Group(Suppress('[') + delimitedList(rval) + Suppress(']'))
        """
        self.value = self.token = tokens[0]
        self.convert()

    def convert(self):
        """ Method to (optionally) convert the input token(s) into something
        else. Particularly used for conversion to base types. """
        self.value = self.token

    def eval(self, context):
        """ Evaluate the object content against a context. Just return the
        stored (and optionally converted) value by default. """
        return self.value

    def is_implication(self):
        """ To be implemented by subclasses to determine if they ultimately are
        implications. """
        return False

    def __str__(self):
        return "%s" % self.token


class EvalBoolean(EvalBase):
    """ Class to evaluate a parsed boolean constant """
    def convert(self):
        logger.debug("Boolean: %s", self.value)
        self.value = self.token == "True"


class EvalNone(EvalBase):
    """ Class to evaluate a parsed none constant """
    def convert(self):
        logger.debug("None: %s", self.value)
        self.value = None


class EvalInteger(EvalBase):
    """ Class to evaluate a parsed integer constant """
    def convert(self):
        logger.debug("Integer: %s", self.token)
        self.value = int(self.token)


class EvalReal(EvalBase):
    """ Class to evaluate a parsed real constant """
    def convert(self):
        logger.debug("Real: %s", self.token)
        self.value = float(self.token)


class EvalString(EvalBase):
    """ Class to evaluate a parsed string constant """
    def convert(self):
        logger.debug("String: %s", self.token)
        self.value = self.token

    def __str__(self):
        return '"%s"' % self.token


class OperatorRegex(object):
    """ A class implementing operators on regular expressions. """
    def __init__(self, string):
        # We use re.search to implement the membership operator (in) in the
        # sense of "matches anywhere in the string". We can use the regex
        # as given for that.
        self.membership_regex = re.compile(string)
        # For equality matches we use re.match which already anchors the
        # matching at the start of the operand but does not require a match all
        # up until the end of it. So we need to add an explicit end-of-line anchor.
        self.equality_regex = re.compile("%s$" % string)
        # NOTE: Take the multiline semantics of re.search vs. re.match into
        # account when looking to change this (although we're not using
        # multiline mode as of now).

    @staticmethod
    def compare_op_impl(function, other):
        """ Implement handling of iterable operands. """
        if isinstance(other, (list, set)):
            for val in other:
                logger.debug("Regular expression match: %s == %s",
                             function, val)
                if function(val):
                    return True
            return False

        if other is None:
            return False

        return function(other)

    def __eq__(self, other):
        """ Implement equality using re.match """
        logger.debug("Regular expression match: %s == %s",
            self.equality_regex, other)
        return self.compare_op_impl(self.equality_regex.match, other)

    def __ne__(self, other):
        """ Implement inequality using re.match """
        logger.debug("Regular expression match: %s != %s",
            self.equality_regex, other)
        return not self.compare_op_impl(self.equality_regex.match, other)

    def __contains__(self, other):
        """ Implement membership using re.search """
        logger.debug("Regular expression search: %s in %s",
            self.membership_regex, other)
        return self.compare_op_impl(self.membership_regex.search, other)


class EvalRegex(EvalBase):
    """ Class to evaluate a regular expression """
    def convert(self):
        logger.debug("Regular expression: %s", self.token)
        self.value = OperatorRegex(self.token)

    def eval(self, context):
        return self.value

    def __str__(self):
        return "/%s/" % self.token


class RegexIterableMixIn(object):
    """ Common functionality for lists and sets containing regular expressions
    with different behaviour of membership operators. """
    def __eq__(self, other):
        if not isinstance(other, (list, set)):
            other = [other]

        # in contrast to normal lists, a list of regexes compared to a list
        # of strings is considered equal if any regex matches any string
        for regex in self:
            logger.debug("Eval regex: %s == %s", regex, other)
            if regex == other:
                return True

        return False

    def __ne__(self, other):
        return not self == other

    def __contains__(self, item):
        for regex in self:
            logger.debug("Eval regex: %s in %s", regex, item)
            # we implement "regex in string" of our grammar as "string in
            # regex" so that our overridden operator
            # regex.__contains__(string) is called and searching can be
            # done. Otherwise error "TypeError: 'in <string>' requires
            # string as left operand, not OperatorRegex" would ensue.
            if item in regex:
                return True

        return False


class RegexList(RegexIterableMixIn, list):
    """ A list containing regular expressions with different behaviour of
    membership operators. """


class RegexSet(RegexIterableMixIn, set):
    """ A set containing regular expressions with different behaviour of
    membership operators. """


class EvalList(EvalBase):
    """ Class to evaluate a parsed list """
    def eval(self, context):
        logger.debug("List: %s", self.value)
        ret = []
        regexes = False
        for val in self.value:
            element = val.eval(context)
            if isinstance(element, OperatorRegex):
                regexes = True
            ret.append(element)
        if regexes:
            return RegexList(ret)
        return ret

    def __str__(self):
        return "[%s]" % (", ".join(["%s" % x for x in self.token]))

class EvalSet(EvalBase):
    """ Class to evaluate a parsed set """
    def eval(self, context):
        logger.debug("Set: %s", self.value)
        ret = set()
        regexes = False
        for val in self.value:
            element = val.eval(context)
            if isinstance(element, OperatorRegex):
                regexes = True
            # For our use case None being an element of a set() is of no use
            # and highly confusing. Supress it
            if element is None:
                continue
            ret.add(element)
        if regexes:
            return RegexSet(ret)
        return ret

    def __str__(self):
        return "{%s}" % (", ".join(["%s" % x for x in self.token]))


class IdentifierMissingException(KeyError):
    def __init__(self, name):
        super().__init__("Identifier '%s' is missing" % name)
        self.name = name


class EvalIdentifier(EvalBase):
    """ Class to evaluate a parsed object name """
    def eval(self, context):
        logger.debug("Identifier: %s", self.value)

        # return the literal identifier name if dereferencing it is disabled
        if isinstance(context, dict) and not context.get('deref', True):
            return self.value

        # potentially raise an actual KeyError here to not mask it as missing
        # identifier
        variables = context['variables']
        try:
            # look the identifier up in the variables part of the context
            return variables[self.value]
        except KeyError as error:
            raise IdentifierMissingException(self.value)


class EvalResult(EvalBase):
    """ Class to evaluate a analysis result """
    def convert(self):
        logger.debug("Result: %s", self.token)
        result_map = {
            'fail': Result.failed,
            'ignore': Result.ignored,
        }

        if self.token in result_map:
            self.value = result_map[self.token]
        else:
            self.value = Result[self.token]


class EvalModifier(EvalBase):
    """ Class to evaluate typical single-operand modifier expressions such as
    explicit sign change, bitwise and logical not. """
    def __init__(self, tokens):
        super().__init__(tokens)
        self.operator, self.value = tokens[0]

    def eval(self, context):
        val = self.value.eval(context)
        if self.operator == '+':
            return val
        elif self.operator == '-':
            return -val
        elif self.operator == '~':
            return ~val
        elif self.operator == 'not':
            return not val

        raise ValueError('Invalid operator %s' % self.operator)

    def __str__(self):
        return "(%s%s)" % (self.operator, self.value)


class EvalPower(EvalBase):
    """ Class to evaluate exponentiation expressions """
    def eval(self, context):
        res = self.value[-1].eval(context)
        for val in self.value[-3::-2]:
            res = val.eval(context)**res

        return res

    def __str__(self):
        return "(%s)" % (" ".join(["%s" % x for x in self.token]))


def operator_operands(tokenlist):
    """ Generator to extract operators and operands in pairs """
    iterator = iter(tokenlist)
    while True:
        try:
            yield (next(iterator), next(iterator))
        except StopIteration:
            break


class EvalArith(EvalBase):
    """ Class to evaluate typical arithmetic and bitwise operations like
    addition, multiplication, division and shifts expressions. Operator
    precedence is handled by the order in which they're evaluated by the
    parser, i.e. given to infixNotation. """
    def eval(self, context):
        ret = self.value[0].eval(context)
        for op, val in operator_operands(self.value[1:]):
            if op == '+':
                ret += val.eval(context)
            elif op == '-':
                ret -= val.eval(context)
            elif op == '*':
                ret *= val.eval(context)
            elif op == '/':
                ret /= val.eval(context)
            elif op == '//':
                ret //= val.eval(context)
            elif op == '%':
                ret %= val.eval(context)
            elif op == '<<':
                ret <<= val.eval(context)
            elif op == '>>':
                ret >>= val.eval(context)
            elif op == '&':
                ret &= val.eval(context)
            elif op == '^':
                ret ^= val.eval(context)
            elif op == '|':
                ret |= val.eval(context)
            elif op == '.':
                # expect op to be an identifier, have it return its name by
                # setting deref to False and then resolve that property in
                # current ret by calling getattr() on it
                property_context = context.copy()
                property_context['deref'] = False
                property_name = val.eval(property_context)
                ret = getattr(ret, property_name)
            elif op == "->":
                if ret:
                    ret = val.eval(context)
                else:
                    ret = None
            else:
                raise ValueError('Invalid operator %s' % op)

        return ret

    def is_implication(self):
        """ Determines if this object is ultimately an implication. """
        try:
            return self.value[-2] == '->'
        except KeyError:
            return False

    def __str__(self):
        return "(%s)" % (" ".join(["%s" % x for x in self.token]))


class EvalLogic(EvalBase):
    """ Class to evaluate comparison expressions """
    def __init__(self, tokens):
        super().__init__(tokens)
        self.operator_map = {
            "<": operator.lt,
            "<=": operator.le,
            ">": operator.gt,
            ">=": operator.ge,
            "==": operator.eq,
            "!=": operator.ne,
            "in": EvalLogic.in_,
            "not in": EvalLogic.not_in,
            "is": operator.is_,
            "is not": operator.is_not,
            "isdisjoint": lambda a, b: a.isdisjoint(b),
            "and": operator.and_,
            "or": operator.or_,
        }

    @staticmethod
    def in_(a, b):
        """ Literally implement membership test. Make it a static method so we
        can do identity checks. Do not use operator.contains because it needs
        operands swapped. """
        return a in b

    @staticmethod
    def not_in(a, b):
        """ Naively implement non-membership test. """
        return a not in b

    @staticmethod
    def handle_regexes(function, val1, val2):
        """ Special handling of equality and membership checks for regular
        expressions. """
        if (function in (operator.eq, operator.ne)
                and isinstance(val2, (OperatorRegex, RegexIterableMixIn))):
            # swap operands around in case the first does not contain any regex
            # but the other does to reliably reroute to our overridden __eq__
            # operator, just do that always to keep checks simple since
            # (in)equality is commutative anyway
            val1, val2 = val2, val1
        elif (function in (EvalLogic.in_, EvalLogic.not_in)
              and isinstance(val1, (OperatorRegex, RegexIterableMixIn))):
            # "<regex> in <string>|<list-of-strings>" of our grammar directly
            # implemented using the "in" operator would call
            # <string>|<list-of-strings>.__contains__(<regex>) which we cannot
            # override with reasonable effort. To get a call of
            # <regex>.__contains__(<string>|<list-of-strings>) we need to
            # switch operands. Otherwise error "TypeError: 'in <string>'
            # requires string as left operand, not OperatorRegex" would ensue.
            val1, val2 = val2, val1

        # nothing special
        return function(val1, val2)

    def eval(self, context):
        val1 = self.value[0].eval(context)
        for op, parseobj in operator_operands(self.value[1:]):
            val2 = parseobj.eval(context)
            logger.debug("Comparison: %s %s %s", val1, op, val2)
            function = self.operator_map[op]
            if not self.handle_regexes(function, val1, val2):
                break
            val1 = val2
        else:
            return True

        return False

    def __str__(self):
        return "(%s)" % (" ".join(["%s" % x for x in self.token]))


class ExpressionParser(object):
    """ Define and run the parser. """
    def __init__(self):
        # speed up infixNotation considerably at the price of some cache memory
        ParserElement.enablePackrat()

        boolean = Keyword('True') | Keyword('False')
        none = Keyword('None')
        integer = Word(nums)
        real = Combine(Word(nums) + "." + Word(nums))
        string = (QuotedString('"', escChar='\\')
                  | QuotedString("'", escChar='\\'))
        regex = QuotedString('/', escChar='\\')
        identifier = Word(alphas, alphanums + '_')
        dereference = infixNotation(identifier, [
            (Literal('.'), 2, opAssoc.LEFT, EvalArith),
        ])
        result = (Keyword('bad') | Keyword('fail') | Keyword('good')
                  | Keyword('ignore') | Keyword('unknown'))
        rval = boolean | none | real | integer | string | regex | result | dereference
        rvallist = Group(Suppress('[') + delimitedList(rval) + Suppress(']'))
        rvalset = Group(Suppress('{') + delimitedList(rval) + Suppress('}'))
        operand = rval | rvallist | rvalset

        # parse actions replace the parsed tokens with an instantiated object
        # which we can later call into for evaluation of its content
        boolean.setParseAction(EvalBoolean)
        none.setParseAction(EvalNone)
        integer.setParseAction(EvalInteger)
        real.setParseAction(EvalReal)
        string.setParseAction(EvalString)
        regex.setParseAction(EvalRegex)
        identifier.setParseAction(EvalIdentifier)
        result.setParseAction(EvalResult)
        rvallist.setParseAction(EvalList)
        rvalset.setParseAction(EvalSet)

        identity_test = Keyword('is') + ~Keyword('not') | Combine(
            Keyword('is') + Keyword('not'), adjacent=False, joinString=' ')
        membership_test = Keyword('in') | Combine(
            Keyword('not') + Keyword('in'), adjacent=False, joinString=' ')
        comparison_op = oneOf('< <= > >= != == isdisjoint')
        comparison = identity_test | membership_test | comparison_op

        self.parser = infixNotation(operand, [
            (Literal('**'), 2, opAssoc.LEFT, EvalPower),
            (oneOf('+ - ~'), 1, opAssoc.RIGHT, EvalModifier),
            (oneOf('* / // %'), 2, opAssoc.LEFT, EvalArith),
            (oneOf('+ -'), 2, opAssoc.LEFT, EvalArith),
            (oneOf('<< >>'), 2, opAssoc.LEFT, EvalArith),
            (Literal('&'), 2, opAssoc.LEFT, EvalArith),
            (Literal('^'), 2, opAssoc.LEFT, EvalArith),
            (Literal('|'), 2, opAssoc.LEFT, EvalArith),
            (comparison, 2, opAssoc.LEFT, EvalLogic),
            (Keyword('not'), 1, opAssoc.RIGHT, EvalModifier),
            (Keyword('and'), 2, opAssoc.LEFT, EvalLogic),
            (Keyword('or'), 2, opAssoc.LEFT, EvalLogic),
            (Keyword('->'), 2, opAssoc.LEFT, EvalArith),
        ])

    def parse(self, expression):
        """ Parse an expression and return an object supporting evaluation of
        that expression against a context. """
        try:
            return self.parser.parseString(expression, parseAll=True)[0]
        except ParseException as parse_error:
            col = parse_error.col
            raise SyntaxError(
                "Expression parse error near character %d: %s>>%s<<%s" % (
                    parse_error.col, expression[0:col], expression[col],
                    expression[col+1:]))


if __name__ == '__main__':
    print(ExpressionParser().parse('foo == (bar - blub)'))



###############################################################################
#                                                                             #
# Peekaboo Extended Email Attachment Behavior Observation Owl                 #
#                                                                             #
# toolbox/                                                                    #
#         files.py                                                            #
###############################################################################
#                                                                             #
# Copyright (C) 2016-2020  science + computing ag                             #
#                                                                             #
# This program is free software: you can redistribute it and/or modify        #
# it under the terms of the GNU General Public License as published by        #
# the Free Software Foundation, either version 3 of the License, or (at       #
# your option) any later version.                                             #
#                                                                             #
# This program is distributed in the hope that it will be useful, but         #
# WITHOUT ANY WARRANTY; without even the implied warranty of                  #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU           #
# General Public License for more details.                                    #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#                                                                             #
###############################################################################


import logging
import mimetypes

import magic


logger = logging.getLogger(__name__)


class Filetools(object):
    """ Parent class, defines interface to various file tools. """
    def __init__(self, sample):
        self.sample = sample

        if not mimetypes.inited:
            mimetypes.init()
            mimetypes.add_type('application/javascript', '.jse')

    def guess_mime_type_from_contents(self):
        """  Get type from file magic bytes. """
        mime_type = magic.from_file(self.sample.file_path, mime=True)
        if not mime_type:
            return None

        return mime_type

    def guess_mime_type_from_name(self):
        """ Guess the type of a file based on its filename or URL. """
        filename = self.sample.name_declared
        if not filename:
            filename = self.sample.filename
        mime_type = mimetypes.guess_type(filename)[0]
        if not mime_type:
            return None

        return mime_type

    def guess_mime_type_text_representation(self):
        """ Guess the type by content and hand back text representation rather
        than mime type. """
        type_as_text = magic.from_file(self.sample.file_path)
        if not type_as_text:
            return None

        return type_as_text

    def get_report(self):
        """ Return filetools report or create if not already cached. """
        if self.sample.filetools_report is not None:
            return self.sample.filetools_report

        ftreport = FiletoolsReport({
            "type_by_content": self.guess_mime_type_from_contents(),
            "type_by_name": self.guess_mime_type_from_name(),
            "type_as_text": self.guess_mime_type_text_representation(),
        })

        self.sample.register_filetools_report(ftreport)
        return ftreport


class FiletoolsReport(object):
    """ Represents a custom Filetools report. """
    def __init__(self, report=None):
        if report is None:
            report = {}
        self.report = report

    def __str__(self):
        return "<FiletoolsReport('%s'>" % self.report

    @property
    def type_by_content(self):
        """ Returns the type of a file guessed based on its contents. """
        return self.report.get('type_by_content', None)

    @property
    def type_by_name(self):
        """ Returns the type of a file guessed based on its filename. """
        return self.report.get('type_by_name', None)

    @property
    def type_as_text(self):
        """ Returns the guessed type of a file in text representation rather
        than mime type. Equal to using the `file` command. """
        return self.report.get('type_as_text', None)

    @property
    def mime_types(self):
        """ Return set of determined mime types. """
        mime_types = {self.type_by_name, self.type_by_content}
        mime_types.discard(None)

        return mime_types



###############################################################################
#                                                                             #
# Peekaboo Extended Email Attachment Behavior Observation Owl                 #
#                                                                             #
# ruleset/                                                                    #
#         __init__.py                                                         #
###############################################################################
#                                                                             #
# Copyright (C) 2016-2020  science + computing ag                             #
#                                                                             #
# This program is free software: you can redistribute it and/or modify        #
# it under the terms of the GNU General Public License as published by        #
# the Free Software Foundation, either version 3 of the License, or (at       #
# your option) any later version.                                             #
#                                                                             #
# This program is distributed in the hope that it will be useful, but         #
# WITHOUT ANY WARRANTY; without even the implied warranty of                  #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU           #
# General Public License for more details.                                    #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#                                                                             #
###############################################################################


from enum import Enum, unique


@unique
class Result(Enum):
    """ Enumeration of rule evaluation result severities with apropriate
    comparison semantics. """
    unchecked = 1
    unknown = 2
    ignored = 3
    good = 4
    failed = 5
    bad = 6

    def __ge__(self, other):
        if self.__class__ is other.__class__:
            return self.value >= other.value
        return NotImplemented

    def __le__(self, other):
        if self.__class__ is other.__class__:
            return self.value <= other.value
        return NotImplemented


class RuleResult:
    """ A rule evaluation result with severity, deciding rule, decision
    description and reason. """
    def __init__(self, rule,
                 result=Result.unknown,
                 reason=None,
                 further_analysis=True):
        self.result = Result.unchecked
        self.rule = rule
        self.result = result
        self.reason = reason
        if self.reason is None:
            self.reason = _("Rule without result")
        self.further_analysis = further_analysis

    def __str__(self):
        return (_("Result \"%s\" of rule %s - %s, analysis continues: %s.")
                % (self.result.name, self.rule, self.reason,
                   _('Yes') if self.further_analysis else _('No')))

    __repr__ = __str__



###############################################################################
#                                                                             #
# Peekaboo Extended Email Attachment Behavior Observation Owl                 #
#                                                                             #
# server.py                                                                   #
###############################################################################
#                                                                             #
# Copyright (C) 2016-2020  science + computing ag                             #
#                                                                             #
# This program is free software: you can redistribute it and/or modify        #
# it under the terms of the GNU General Public License as published by        #
# the Free Software Foundation, either version 3 of the License, or (at       #
# your option) any later version.                                             #
#                                                                             #
# This program is distributed in the hope that it will be useful, but         #
# WITHOUT ANY WARRANTY; without even the implied warranty of                  #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU           #
# General Public License for more details.                                    #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#                                                                             #
###############################################################################

""" This module implements the Peekaboo server, i.e. the frontend to the
client. """

import errno
import grp
import json
import logging
import os
import stat
import socket
import socketserver
from threading import Thread, Event, current_thread
from peekaboo.ruleset import Result


logger = logging.getLogger(__name__)


class PeekabooStreamServer(socketserver.ThreadingUnixStreamServer):
    """ Asynchronous server. """
    def __init__(self, server_address, request_handler_cls, job_queue,
                 sample_factory, bind_and_activate=True,
                 request_queue_size=10, status_change_timeout=60):
        self.server_address = server_address
        self.__job_queue = job_queue
        self.__sample_factory = sample_factory
        self.request_queue_size = request_queue_size
        self.allow_reuse_address = True
        self.status_change_timeout = status_change_timeout
        self.__shutdown_requested = False
        self.__request_triggers = {}

        # no super() since old-style classes
        logger.debug('Starting up server.')
        socketserver.ThreadingUnixStreamServer.__init__(
            self, server_address, request_handler_cls,
            bind_and_activate=bind_and_activate)

    @property
    def job_queue(self):
        """ Return this server's reference to the job queue. Used by handler
        threads to get access to it for submission of samples for processing.
        """
        return self.__job_queue

    @property
    def sample_factory(self):
        """ Return this server's reference to a factory that can create
        pre-configured sample objects. Used by handler threads to get access to
        it for creation of samples prior to submission for processing. """
        return self.__sample_factory

    @property
    def shutting_down(self):
        """ Return True if we've received a shutdown request. """
        return self.__shutdown_requested

    def register_request(self, thread, event):
        """ Register an event for a request being handled to trigger if we want
        it to shut down. """
        self.__request_triggers[thread] = event
        logger.debug('Request registered with server.')

    def deregister_request(self, thread):
        """ Deregister a request which has finished handling and does no logner
        need to be made aware that we want it to shut down. """
        logger.debug('Request deregistered from server.')
        del self.__request_triggers[thread]

    def shutdown(self):
        """ Shut down the server. In our case, notify requests which are
        currently being handled to shut down as well. """
        logger.debug('Server shutting down.')
        self.__shutdown_requested = True
        for thread in self.__request_triggers:
            # wake up the thread so it can see that we're shutting down
            self.__request_triggers[thread].set()

        socketserver.ThreadingUnixStreamServer.shutdown(self)

    def server_close(self):
        """ Finally completely close down the server. """
        # no new connections from this point on
        logger.debug('Removing connection socket %s', self.server_address)
        try:
            os.remove(self.server_address)
        except OSError as oserror:
            logger.warning('Removal of socket %s failed: %s',
                           self.server_address, oserror)

        logger.debug('Closing down server.')
        return socketserver.ThreadingUnixStreamServer.server_close(self)


class PeekabooStreamRequestHandler(socketserver.StreamRequestHandler):
    """ Request handler used by PeekabooStreamServer to handle analysis
    requests. """
    def setup(self):
        # rename thread for higher log message clarity
        thread = current_thread()
        # keep trailing thread number by replacing just the base name
        thread.name = thread.name.replace('Thread-', 'Request-')

        socketserver.StreamRequestHandler.setup(self)
        self.job_queue = self.server.job_queue
        self.sample_factory = self.server.sample_factory
        self.status_change_timeout = self.server.status_change_timeout

        # create an event we will give to all the samples and our server to
        # wake us if they need out attention
        self.status_change = Event()
        self.status_change.clear()

    def handle(self):
        """ Handles an analysis request. """
        # catch wavering clients early on
        logger.debug('New connection incoming.')
        if not self.talk_back([_('Hello, this is Peekaboo.'), '']):
            return

        # parse client input and see if it's valid and whether there are any
        # requests to us in there
        requests = self.parse()
        if not requests:
            return

        # handle client requests. Sample submissions are special in that they
        # submit into our work queue and we need to wait for them to finish. So
        # see if handling the requests did submit something.
        submitted = self.handle_requests(requests)
        if not submitted:
            return

        # wait for jobs submitted by request handling
        if not self.wait(submitted):
            # something went wrong while waiting, i.e. client closed connection
            # or we're shutting down
            return

        # here we know that all samples have reported back
        self.report(submitted)

    def submit_sample(self, api_data):
        """ Submit API supplied file as Sample """
        path = api_data['full_name']
        logger.info("Got run_analysis request for %s", path)
        if not os.path.exists(path):
            self.talk_back(_('ERROR: Path does not exist or no '
                             'permission to access it.'))
            logger.error('Path does not exist or no permission '
                         'to access it.')
            return None

        if not os.path.isfile(path):
            self.talk_back(_('ERRROR: Input is not a file'))
            logger.error('Input is not a file')
            return None

        sample = self.sample_factory.make_sample(
            path, status_change=self.status_change, metainfo=api_data)
        if not self.job_queue.submit(sample, self.__class__):
            self.talk_back(_('Error submitting sample to job queue'))
            # submit will have logged an error
            return None
        return sample

    def parse(self):
        """ Reads and parses an analysis request. This is expected to be a JSON
        structure::

            [ { "full_name": "<path>",
                "name_declared": ...,
                ... },
              { "request": "scan-file",
                "full_name": ... },
              { "request": "ping" },
              ... ]

        The request field is optional with a default of "scan-file".
        Valid requests are:
        * scan-file
        * ping

        scan-file: requires the path of the directory / file to analyse
        in "full_name".

        The maximum buffer size is 16 KiB, because JSON incurs some bloat.
        """
        try:
            client_input = self.request.recv(1024 * 16).rstrip()
        except IOError as ioerror:
            logger.error('Request receive failed: %s', ioerror)
            return None

        try:
            requests = json.loads(client_input)
        except ValueError as error:
            self.talk_back(_('Error: Invalid JSON in request.'))
            logger.error('Invalid JSON in request: %s', error)
            return None

        # do some generic input validation
        if not isinstance(requests, (list, tuple)):
            self.talk_back(_('ERROR: Invalid data structure.'))
            logger.error('Invalid data structure.')
            return None

        for request in requests:
            if not isinstance(request, dict):
                self.talk_back(_('ERROR: Invalid request structure.'))
                logger.error('Invalid request structure.')
                return None

            # default request type is scan-file
            reqtype = request.get('request', 'scan-file')
            if reqtype not in ['scan-file', 'ping']:
                self.talk_back(_('ERROR: Unsupported request %s') % reqtype)
                logger.error('Unsupported request %s', reqtype)
                return None

            if reqtype == 'scan-file' and 'full_name' not in request:
                self.talk_back(_('ERROR: Incomplete data structure.'))
                logger.error('Incomplete data structure.')
                return None

        return requests

    def handle_requests(self, requests):
        """ Handle requests.

        @param requests: list/tuple of dicts describing requests
        @returns: list of submitted samples or None on error
        """
        submitted = []
        for request in requests:
            reqtype = request.get('request', 'scan-file')
            if reqtype == 'scan-file':
                sample = self.submit_sample(request)
                if sample is None:
                    # here we abort the whole batch of requests which is
                    # somewhat excessive since the others might just be pings
                    # or complete successfully. But since we can't fulfill all
                    # the requested analyses and have no way of communicating
                    # that fact individually to the client it's the only option
                    # for now.
                    return None

                submitted.append(sample)
                logger.debug('Created and submitted sample %s', sample)
            elif reqtype == 'ping':
                self.talk_back(json.dumps({'request': 'ping',
                                           'response': 'pong'}))
                logger.debug('ping request received and answered')
            else:
                # paranoia
                logger.error('Unknown request "%s" got through input '
                             'validation', reqtype)
                self.talk_back(_('ERROR: Unsupported request %s') % reqtype)
                return None

        return submitted

    def wait(self, to_be_analysed):
        """ Wait for submitted analysis jobs to finished.

        @param to_be_analysed: samples that have been submitted for analysis
                               and which will report back to us when they're
                               finished.
        @type to_be_analysed: List of Sample objects
        """
        # register with our server so it can notify us if it wants us to shut
        # down
        # NOTE: Every exit point from this routine needs to deregister this
        # request from the server to avoid memory leaks. Unfortunately, the
        # server cannot do this iteself on shutdown_request() because it does
        # not have any thread ID available there.
        self.server.register_request(self, self.status_change)

        # wait for results to come in
        while to_be_analysed:
            # wait for an event to signal that its status has changed or
            # timeout expires
            if not self.status_change.wait(self.status_change_timeout):
                # keep our client engaged
                # TODO: Impose maximum processing time of our own?
                if not self.talk_back(_('Files are being analyzed...')):
                    # Abort handling this request since no-one's interested
                    # any more. We could dequeue the samples here to avoid
                    # unnecessary work. Instead we'll have them run their
                    # course, assuming that we'll be much quicker
                    # responding if the client decides to resubmit them.
                    self.server.deregister_request(self)
                    return False

                logger.debug('Client updated that samples are still '
                             'processing.')

                # Fall through here and evaluate all samples for paranoia's
                # sake in case our status change event has a race condition.
                # It shouldn't though, because we wait for it, then first clear
                # it and then look at all samples that might have set it. If
                # while doing that another sample sets it and we don't catch it
                # because we've already looked at it, the next wait will
                # immediately return and send us back into checking all samples
                # for status change.

            # see if our server is shutting down and follow it if so
            if self.server.shutting_down:
                self.talk_back(_('Peekaboo is shutting down.'))
                logger.debug('Request shutting down with server.')
                self.server.deregister_request(self)
                return False

            self.status_change.clear()

            # see which samples are done and which are still processing
            still_analysing = []
            for sample in to_be_analysed:
                # remove samples that are done
                if sample.done:
                    continue

                still_analysing.append(sample)

            to_be_analysed = still_analysing

        # deregister notification from server since we've exited our wait loop
        self.server.deregister_request(self)
        return True

    def report(self, done):
        """ Report individual files' and overall verdict to client.

        @param done: List of samples that are done processing and need
                     reporting.
        @type done: List of Sample objects.
        """
        # evaluate results into an overall result: We want to present the
        # client with an overall result instead of confusing them with
        # assertions about individual files. Particularly in the case of
        # AMaViS, this would otherwise lead to messages being passed on as
        # clean where a single attachment evaluated to "good" but analysis of
        # all the others failed.
        result = Result.unchecked
        logger.debug('Determining final verdict to report to client.')
        for sample in done:
            # check if result of this rule is worse than what we know so far
            sample_result = sample.result
            logger.debug('Current overall result: %s, Sample result: %s',
                         result.name, sample_result.name)
            if sample_result >= result:
                result = sample_result

            # and unconditionally send out its report to the client (plus an
            # empty line)
            if not self.talk_back(sample.peekaboo_report + ['']):
                return

        # report overall result.
        logger.debug('Reporting batch as "%s" to client', result.name)
        loc_verdict = _('The file collection has been categorized "%s"')
        overall = [loc_verdict % result.name]

        # Add untranslated verdict (if the above actually got translated) for
        # potential pattern matching of the client to reliably latch on to.
        # Need to duplicate strings here for pygettext and pybabel extract to
        # find the translatable one in the above _().
        verdict = 'The file collection has been categorized "%s"'
        if verdict != loc_verdict:
            overall.append(verdict % result.name)

        # append newline and send
        overall.append('')
        if not self.talk_back(overall):
            return

        # shut down connection
        logger.debug('Results reported back to client - closing connection.')

    def talk_back(self, msgs):
        """ Send message(s) back to the client. Automatically appends newline
        to each message.

        @param msgs: message(s) to send to client.
        @type msgs: string or (list or tuple of strings)

        @returns: True on successful sending of all messages, False on error of
                  sending and None specifically if sending failed because the
                  client closed the connection. """
        if not isinstance(msgs, (list, tuple)):
            msgs = (msgs, )

        for msg in msgs:
            try:
                # FIXME: Hard-coded, arbitrary encoding since we have no
                # clearly defined protocol here.
                self.request.sendall(('%s\n' % msg).encode('utf-8'))
            except IOError as ioerror:
                if ioerror.errno == errno.EPIPE:
                    logger.warning('Client closed connection on us: %s',
                                   ioerror)
                    return None

                logger.warning('Error talking back to client: %s', ioerror)
                return False

        return True


class PeekabooServer(object):
    """ A class wrapping the server components of Peekaboo. """
    def __init__(self, sock_file, job_queue, sample_factory, request_queue_size,
                 sock_group = None,
                 sock_mode = stat.S_IREAD | stat.S_IWRITE |
                             stat.S_IRGRP | stat.S_IWGRP |
                             stat.S_IROTH | stat.S_IWOTH):

        """ Initialise a new server and start it. All error conditions are
        returned as exceptions.

        @param sock_file: The path of the socket file.
        @type sock_file: String
        @param job_queue: A reference to the job queue for submission of
                          samples.
        @type job_queue: JobQueue
        @param sample_factory: A reference to a sample factory for creating new
                               samples.
        @type sample_factory: SampleFactory
        @param request_queue_size: Number of requests that may be pending on
                                   the socket.
        @type request_queue_size: int
        @param sock_group: An optional name of a group that shall own the
                           socket.
        @type sock_group: String
        @param sock_mode: The permission bits of the socket.
        @type sock_mode: Integer (bitmask)
        """
        self.server = None
        self.runner = None

        self.server = PeekabooStreamServer(
            sock_file,
            PeekabooStreamRequestHandler,
            job_queue=job_queue,
            sample_factory=sample_factory,
            request_queue_size=request_queue_size)

        self.runner = Thread(target=self.server.serve_forever, name="Server")
        self.runner.start()

        try:
            if sock_mode is not None:
                os.chmod(sock_file, sock_mode)

            if sock_group is not None:
                gid = grp.getgrnam(sock_group)[2]
                os.chown(sock_file, -1, gid)
        except:
            self.shutdown()
            raise

        logger.info('Peekaboo server is now listening on %s',
                    self.server.server_address)

    def shutdown(self):
        """ Shuts down the server. """
        self.server.shutdown()
        self.server.server_close()
        self.runner.join()



###############################################################################
#                                                                             #
# Peekaboo Extended Email Attachment Behavior Observation Owl                 #
#                                                                             #
# db.py                                                                       #
###############################################################################
#                                                                             #
# Copyright (C) 2016-2020  science + computing ag                             #
#                                                                             #
# This program is free software: you can redistribute it and/or modify        #
# it under the terms of the GNU General Public License as published by        #
# the Free Software Foundation, either version 3 of the License, or (at       #
# your option) any later version.                                             #
#                                                                             #
# This program is distributed in the hope that it will be useful, but         #
# WITHOUT ANY WARRANTY; without even the implied warranty of                  #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU           #
# General Public License for more details.                                    #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#                                                                             #
###############################################################################

""" A class wrapping database operations needed by Peekaboo based on
SQLAlchemy. """

import threading
import logging
from datetime import datetime, timedelta
import sqlalchemy
from sqlalchemy import Column, Integer, String, Text, DateTime, ForeignKey, \
        Enum, Index
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.engine import create_engine
from sqlalchemy.orm import sessionmaker, scoped_session, relationship
from sqlalchemy.exc import SQLAlchemyError, IntegrityError
from peekaboo import __version__
from peekaboo.ruleset import Result
from peekaboo.exceptions import PeekabooDatabaseError

DB_SCHEMA_VERSION = 7

logger = logging.getLogger(__name__)
Base = declarative_base()


#
# Database schema definition.
##############################################################################


class PeekabooMetadata(Base):
    """ Definition of the _meta table for Peekaboo. """
    __tablename__ = '_meta'
    peekaboo_version = Column(String(10), nullable=False,
                              primary_key=True)
    db_schema_version = Column(Integer, nullable=False,
                               primary_key=True,
                               autoincrement=False)
    cuckoo_version = Column(String(10), nullable=False,
                            primary_key=True)

    def __str__(self):
        return (
            '<PeekabooMetadata(peekaboo_version="%s", db_schema_version="%s", '
            'cuckoo_version="%s")>'
            % (self.peekaboo_version,
               self.db_schema_version,
               self.cuckoo_version)
        )

    __repr__ = __str__


class InFlightSample(Base):
    """
    Table tracking whether a specific sample is currently being analysed and by
    which Peekaboo instance.
    """
    __tablename__ = 'in_flight_samples_v%d' % DB_SCHEMA_VERSION

    # Indices:
    # - general considerations: The table will likely never have more than a
    #   couple of hundret entries. But we add and delete quite frequently.
    # - uniqueness of the primary key ensures atomic insertion when adding a
    #   lock
    # - column: we delete our own stale locks by instance_id.
    # - column: we delete other's stale locks by start_time.
    # - compound: we delete our own locks by sha256sum and instance_id.
    #   (admittedly a bit of overkill since the individual columns are already
    #   indexed.)

    sha256sum = Column(String(64), primary_key=True)
    instance_id = Column(Integer, nullable=False, index=True)
    start_time = Column(DateTime, nullable=False, index=True)

    __table_args__ = (
        # Index names need to be unique per schema in postgresql.
        Index('ix_%s_sha_iid' % __tablename__, sha256sum, instance_id),
        )

    def __str__(self):
        return (
            '<InFlightSample(sha256sum="%s", instance_id="%s", '
            'start_time="%s")>'
            % (self.sha256sum,
               self.instance_id,
               self.start_time.strftime("%Y%m%dT%H%M%S"))
        )

    __repr__ = __str__


class SampleInfo(Base):
    """ Definition of the sample_info table. """
    __tablename__ = 'sample_info_v%d' % DB_SCHEMA_VERSION

    # Indices:
    # - general considerations: The table grows very large over time. Every
    #   sample is checked against it to find a cached analysis result.
    #   Otherwise it's quite unused currently.
    # - compound: we frequently search by sha256sum and file extension

    id = Column(Integer, primary_key=True)
    sha256sum = Column(String(64), nullable=False)
    file_extension = Column(String(16), nullable=True)
    result = Column(Enum(Result), nullable=False)
    reason = Column(Text, nullable=True)

    __table_args__ = (
        Index('ix_%s_sha_fe' % __tablename__, sha256sum, file_extension),
        )

    def __str__(self):
        return ('<SampleInfo(sample_sha256_hash="%s", file_extension="%s", '
                'reason="%s")>'
                % (self.sha256sum,
                   self.file_extension,
                   self.reason))

    __repr__ = __str__


class AnalysisJournal(Base):
    """ Definition of the analysis_jobs table. """
    __tablename__ = 'analysis_jobs_v%d' % DB_SCHEMA_VERSION

    # Indices:
    # - general considerations: We're not using this table at all currently.
    #   All indexes are therefore currenly only futureproofing and admin or
    #   scripting aids.
    # - column: Speed up removal of old entries by analysis_time.
    # - column: Index foreign key to speed up search and removal by sample id.

    id = Column(Integer, primary_key=True)
    job_hash = Column(String(255), nullable=False)
    cuckoo_job_id = Column(Integer, nullable=False)
    filename = Column(String(255), nullable=False)
    analysis_time = Column(DateTime, nullable=False,
                           index=True)
    sample_id = Column(Integer, ForeignKey(SampleInfo.id),
                       nullable=False, index=True)
    sample = relationship('SampleInfo')

    def __str__(self):
        return (
            '<AnalysisJournal(job_hash="%s", cuckoo_job_id="%s", '
            'filename="%s", analysis_time="%s")>'
            % (self.job_hash,
               self.cuckoo_job_id,
               self.filename,
               self.analysis_time.strftime("%Y%m%dT%H%M%S"))
        )

    __repr__ = __str__


#
# End of database schema definition.
##############################################################################


class PeekabooDatabase(object):
    """ Peekaboo's database. """
    def __init__(self, db_url, instance_id=0,
                 stale_in_flight_threshold=15*60,
                 log_level=logging.WARNING):
        """
        Initialize the Peekaboo database handler.

        @param db_url: An RFC 1738 URL that points to the database.
        @param instance_id: A positive, unique ID differentiating this Peekaboo
                            instance from any other instance using the same
                            database for concurrency coordination. Value of 0
                            means that we're alone and have no other instances
                            to worry about.
        @param stale_in_flight_threshold: Number of seconds after which a in
        flight marker is considered stale and deleted or ignored.
        """
        logging.getLogger('sqlalchemy.engine').setLevel(log_level)

        self.__engine = create_engine(db_url, pool_recycle=1)
        session_factory = sessionmaker(bind=self.__engine)
        self.__session = scoped_session(session_factory)
        self.__lock = threading.RLock()
        self.instance_id = instance_id
        self.stale_in_flight_threshold = stale_in_flight_threshold
        if not self._db_schema_exists():
            self._init_db()
            logger.debug('Database schema created.')

    def analysis_save(self, sample):
        """
        Save an analysis task to the analysis journal in the database.

        @param sample: The sample object for this analysis task.
        """
        analysis = AnalysisJournal()
        analysis.job_hash = sample.job_hash
        analysis.cuckoo_job_id = sample.job_id
        analysis.filename = sample.filename
        analysis.analysis_time = datetime.now()
        sample_info = self.sample_info_fetch(sample)
        if sample_info is None:
            sample_info = SampleInfo(
                sha256sum=sample.sha256sum,
                file_extension=sample.file_extension,
                result=sample.result,
                reason=sample.reason)

        analysis.sample = sample_info

        with self.__lock:
            session = self.__session()
            session.add(analysis)
            try:
                session.commit()
            except SQLAlchemyError as error:
                session.rollback()
                raise PeekabooDatabaseError(
                    'Failed to add analysis task to the database: %s' % error
                )
            finally:
                session.close()

    def sample_info_fetch(self, sample):
        """
        Fetch information stored in the database about a given sample object.

        @param sample: The sample object of which the information shall be
                       fetched from the database.
        @return: A SampleInfo object containing the information stored in teh
                 database about the sample.
        """
        with self.__lock:
            session = self.__session()
            sample_info = session.query(SampleInfo).filter_by(
                sha256sum=sample.sha256sum,
                file_extension=sample.file_extension).first()
            session.close()
        return sample_info

    def mark_sample_in_flight(self, sample, instance_id=None, start_time=None):
        """
        Mark a sample as in flight, i.e. being worked on by an instance.

        @param sample: The sample to mark as in flight.
        @param instance_id: (optionally) The ID of the instance that is
                            handling this sample. Default: Us.
        @param start_time: Override the time the marker was placed for
                           debugging purposes.
        """
        # an instance id of 0 denotes that we're alone and don't need to track
        # in-flight samples in the database
        if self.instance_id == 0:
            return True

        # use our own instance id if none is given
        if instance_id is None:
            instance_id = self.instance_id

        if start_time is None:
            start_time = datetime.utcnow()

        session = self.__session()

        # try to mark this sample as in flight in an atomic insert operation
        sha256sum = sample.sha256sum
        session.add(InFlightSample(sha256sum=sha256sum,
                                   instance_id=instance_id,
                                   start_time=start_time))

        locked = False
        try:
            session.commit()
            locked = True
            logger.debug('Marked sample %s as in flight', sha256sum)
        # duplicate primary key == entry already exists
        except IntegrityError:
            session.rollback()
            logger.debug('Sample %s is already in flight on another instance',
                         sha256sum)
        except SQLAlchemyError as error:
            session.rollback()
            raise PeekabooDatabaseError('Unable to mark sample as in flight: '
                                        '%s' % error)
        finally:
            session.close()

        return locked

    def clear_sample_in_flight(self, sample, instance_id=None):
        """
        Clear the mark that a sample is being processed by an instance.

        @param sample: The sample to clear from in-flight list.
        @param instance_id: (optionally) The ID of the instance that is
                            handling this sample. Default: Us.
        """
        # an instance id of 0 denotes that we're alone and don't need to track
        # in-flight samples in the database
        if self.instance_id == 0:
            return

        # use our own instance id if none is given
        if instance_id is None:
            instance_id = self.instance_id

        session = self.__session()

        # clear in-flight marker from database
        sha256sum = sample.sha256sum
        query = session.query(InFlightSample).filter(
            InFlightSample.sha256sum == sha256sum).filter(
                InFlightSample.instance_id == instance_id)

        try:
            # delete() is not queued and goes to the DB before commit()
            cleared = query.delete()
            session.commit()
        except SQLAlchemyError as error:
            session.rollback()
            raise PeekabooDatabaseError('Unable to clear in-flight status of '
                                        'sample: %s' % error)
        finally:
            session.close()

        if cleared == 0:
            raise PeekabooDatabaseError('Unexpected inconsistency: Sample %s '
                                        'not recoreded as in-flight upon '
                                        'clearing flag.' % sha256sum)
        elif cleared > 1:
            raise PeekabooDatabaseError('Unexpected inconsistency: Multiple '
                                        'instances of sample %s in-flight '
                                        'status cleared against database '
                                        'constraints!?' % sha256sum)

        logger.debug('Cleared sample %s from in-flight list', sha256sum)

    def clear_in_flight_samples(self, instance_id=None):
        """
        Clear all in-flight markers left over by previous runs or other
        instances by removing them from the lock table.

        @param instance_id: Clear our own (None), another instance's (positive
                            integer) or all instances' (negative integer) locks.
                            Since an instance_id of 0 disables in-flight sample
                            tracking, no instance will ever set a marker with
                            that ID so that specifying 0 here will amount to a
                            no-op or rather clean-up of invalid entries.
        """
        # an instance id of 0 denotes that we're alone and don't need to track
        # in-flight samples
        if self.instance_id == 0:
            return

        # use our own instance id if none is given
        if instance_id is None:
            instance_id = self.instance_id

        session = self.__session()

        if instance_id < 0:
            # delete all locks
            query = session.query(InFlightSample)
            logger.debug('Clearing database of all in-flight samples.')
        else:
            # delete only the locks of a specific instance
            query = session.query(InFlightSample).filter(
                InFlightSample.instance_id == instance_id)
            logger.debug('Clearing database of all in-flight samples of '
                         'instance %d.', instance_id)
        try:
            # delete() is not queued and goes to the DB before commit()
            query.delete()
            session.commit()
        except SQLAlchemyError as error:
            session.rollback()
            raise PeekabooDatabaseError('Unable to clear the database of '
                                        'in-flight samples: %s' % error)
        finally:
            session.close()

    def clear_stale_in_flight_samples(self):
        """
        Clear all in-flight markers that are too old and therefore stale. This
        detects instances which are locked up, crashed or shut down.
        """
        # an instance id of 0 denotes that we're alone and don't need to track
        # in-flight samples in the database
        if self.instance_id == 0:
            return True

        session = self.__session()

        # delete only the locks of a specific instance
        query = session.query(InFlightSample).filter(
            InFlightSample.start_time <= datetime.utcnow() - timedelta(
                seconds=self.stale_in_flight_threshold))
        logger.debug(
            'Clearing database of all stale in-flight samples '
            '(%d seconds)', self.stale_in_flight_threshold)

        try:
            # the loop triggers the query, so only do it if debugging is
            # enabled
            if logger.isEnabledFor(logging.DEBUG):
                # obviously there's a race between logging and actual delete
                # here, use with caution, compare with actual number of markers
                # cleared below before relying on it for debugging
                for stale in query:
                    logger.debug('Stale in-flight marker to clear: %s', stale)

            # delete() is not queued and goes to the DB before commit()
            cleared = query.delete()
            session.commit()
            if cleared > 0:
                logger.warn('%d stale in-flight samples cleared.', cleared)
        except SQLAlchemyError as error:
            session.rollback()
            raise PeekabooDatabaseError('Unable to clear the database of '
                                        'stale in-flight samples: %s' % error)
        finally:
            session.close()

        return cleared > 0

    def drop(self):
        """ Drop all tables of the database. """
        try:
            Base.metadata.drop_all(self.__engine)
        except SQLAlchemyError as error:
            raise PeekabooDatabaseError(
                'Unable to drop all tables of the database: %s' % error)

    def _db_schema_exists(self):
        insp = sqlalchemy.inspect(self.__engine)
        if not insp.has_table('_meta'):
            return False

        session = self.__session()
        # get the first of potentially multiple metadata entries ordered by
        # descending schema version to get the newest currently configured
        # schema
        meta = session.query(PeekabooMetadata).order_by(
            PeekabooMetadata.db_schema_version.desc()).first()
        if not meta:
            return False

        schema_version = meta.db_schema_version
        session.close()
        if schema_version < DB_SCHEMA_VERSION:
            logger.info('Adding new database schema.')
            return False

        return True

    def _init_db(self):
        """
        Initializes the Peekaboo database by creating tables and
        writing meta information to the '_meta' table.
        """
        Base.metadata.create_all(self.__engine)
        meta = PeekabooMetadata()
        meta.peekaboo_version = __version__
        meta.db_schema_version = DB_SCHEMA_VERSION
        # TODO: Get Cuckoo version.
        meta.cuckoo_version = '2.0'
        session = self.__session()
        session.add(meta)
        try:
            session.commit()
        except SQLAlchemyError as error:
            session.rollback()
            raise PeekabooDatabaseError(
                'Cannot initialize the database: %s' % error)
        finally:
            session.close()



###############################################################################
#                                                                             #
# Peekaboo Extended Email Attachment Behavior Observation Owl                 #
#                                                                             #
# ruleset/                                                                    #
#         engine.py                                                           #
###############################################################################
#                                                                             #
# Copyright (C) 2016-2020  science + computing ag                             #
#                                                                             #
# This program is free software: you can redistribute it and/or modify        #
# it under the terms of the GNU General Public License as published by        #
# the Free Software Foundation, either version 3 of the License, or (at       #
# your option) any later version.                                             #
#                                                                             #
# This program is distributed in the hope that it will be useful, but         #
# WITHOUT ANY WARRANTY; without even the implied warranty of                  #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU           #
# General Public License for more details.                                    #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#                                                                             #
###############################################################################


import logging
from peekaboo.ruleset import Result, RuleResult
from peekaboo.ruleset.rules import *
from peekaboo.toolbox.peekabooyar import ContainsPeekabooYarRule
from peekaboo.exceptions import PeekabooAnalysisDeferred, \
        PeekabooConfigException, PeekabooRulesetConfigError


logger = logging.getLogger(__name__)


class RulesetEngine(object):
    """
    Peekaboo's ruleset engine.

    @since: 1.6
    """
    known_rules = [
        KnownRule,
        FileLargerThanRule,
        ExpressionRule,
        FileTypeOnWhitelistRule,
        FileTypeOnGreylistRule,
        CuckooEvilSigRule,
        CuckooScoreRule,
        OfficeMacroRule,
        OfficeMacroWithSuspiciousKeyword,
        RequestsEvilDomainRule,
        CuckooAnalysisFailedRule,
        ContainsPeekabooYarRule,
        FinalRule
    ]

    def __init__(self, config, db_con):
        """ Initialise the engine, validate its and the individual rules'
        configuration.

        @param config: ruleset configuration parser
        @param db_con: database connection handed to rules (not used by engine
                       itself)
        @raises PeekabooRulesetConfigError: if configuration errors are found
        """
        # create a lookup table from rule name to class
        rule_classes = {}
        for known_rule in self.known_rules:
            rule_classes[known_rule.rule_name] = known_rule

        try:
            enabled_rules = config.getlist('rules', 'rule')
        except PeekabooConfigException as error:
            raise PeekabooRulesetConfigError(
                'Ruleset configuration error: %s' % error)

        if not enabled_rules:
            raise PeekabooRulesetConfigError(
                'No enabled rules found, check ruleset config.')

        # check if unknown rules are enabled
        known_rule_names = rule_classes.keys()
        unknown_rules = set(enabled_rules) - set(known_rule_names)
        if unknown_rules:
            raise PeekabooRulesetConfigError(
                'Unknown rule(s) enabled: %s' % ', '.join(unknown_rules))

        # check for unknown config sections by using rule names as rules'
        # config section names. Allow all known rules not only the enabled ones
        # because some might be temporarily disabled but should be allowed to
        # retain their configuration. Use += extension of list to avoid
        # 'TypeError: can only concatenate list (not "dict_keys") to list' with
        # python3.
        known_sections = ['rules']
        known_sections += known_rule_names
        config.check_sections(known_sections)

        # instantiate enabled rules and have them check their configuration,
        # user-defined rule order is preserved in enabled_rules and through
        # ordered append() in self.rules
        self.rules = []
        for rule in enabled_rules:
            rule = rule_classes[rule](config, db_con)
            self.rules.append(rule)

    def run(self, sample):
        """ Run all the rules in the ruleset against a given sample

        @param sample: sample to evaluate ruleset against
        @returns: Nothing, all state is recorded in the sample """
        for rule in self.rules:
            rule_name = rule.rule_name
            logger.debug("Processing rule '%s' for %s", rule_name, sample)

            try:
                result = rule.evaluate(sample)
                sample.add_rule_result(result)
            except PeekabooAnalysisDeferred:
                # in case the Sample is requesting the Cuckoo report
                raise
            # catch all other exceptions for this rule
            except Exception as error:
                logger.warning("Unexpected error in '%s' for %s", rule_name,
                               sample)
                logger.exception(error)
                # create "fake" RuleResult
                result = RuleResult("RulesetEngine", result=Result.failed,
                                    reason=_("Rule aborted with error"),
                                    further_analysis=False)
                sample.add_rule_result(result)

            logger.info("Rule '%s' processed for %s", rule_name, sample)
            if not result.further_analysis:
                return

        logger.info("Rules evaluated")



###############################################################################
#                                                                             #
# Peekaboo Extended Email Attachment Behavior Observation Owl                 #
#                                                                             #
# __init__.py                                                                 #
###############################################################################
#                                                                             #
# Copyright (C) 2016-2020  science + computing ag                             #
#                                                                             #
# This program is free software: you can redistribute it and/or modify        #
# it under the terms of the GNU General Public License as published by        #
# the Free Software Foundation, either version 3 of the License, or (at       #
# your option) any later version.                                             #
#                                                                             #
# This program is distributed in the hope that it will be useful, but         #
# WITHOUT ANY WARRANTY; without even the implied warranty of                  #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU           #
# General Public License for more details.                                    #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#                                                                             #
###############################################################################

""" Peekaboo constant module data. """


VERSION = (2, 0, 2)

__version__ = '.'.join(map(str, VERSION))
__author__ = 'Felix Bauer'
__description__ = 'Peekaboo Extended Email Attachment Behavior Observation Owl'
__copyright__ = (
    'Copyright (C) 2016-2020 science + computing ag. All rights reserved.')
__license__ = 'GPLv3'


PEEKABOO_OWL = r"""
Peekaboo Extended Email Attachment Behavior Observation Owl

                   _a_aa                    a_aa,
                    '*U4UUUULa_aa_aa_aajUUU4XU7'
                      aX''''''UUXU4XUU'''''!Ua
                    _U'        -U4UU'   _    'U,
                    ?i   jLd1   ?#Wi   4L01   Ui
                    -U,        4#000P        _U'
                     -*Xa_a_a_WUW##KUL_a_a_aX7'
                    _aXUXUUU4UUX4XX444UUUUUUXLa,
                   _UXXUXUXU47'!'!'!'!*X444U4UXX,
                   ?XU4U4''__    ___   ___'UUXUUi
                   ?4U4' |__ \  / _ \ |__ \ 'UUXi
                    *Xi    _) || | | |  _) | ?X7
                     *L   / _/ | |_| | / _/   j7
                      *a |____(_)___(_)____| jY
                       -L,                _/'
                         'l,            _/'
                           j7_a_;  aaa/4
               _aaaaaa#0000#00000##0##00000000aaaaaa,
        aaad0P!!!!!!      Blakiston's Fish       '!!!!!!Laaa
  _aa!!!!                                                    !! _,
(never mind the K)
"""
